# MIT License
#
# Copyright (c) 2023 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import smtpd
import uuid

from pinkman.model.message import Message


class Mailbox(smtpd.SMTPServer):
    """Mailbox Module"""

    def __init__(self, configs, database):
        super().__init__(
            (configs["server"]["hostname"], configs["server"]["port"]), None
        )
        self.configs = configs
        self.database = database

    def process_message(
        self, peer, mailfrom, rcpttos, data, mail_options=[], rcpt_options=[]
    ):
        message = Message(
            str(uuid.uuid4()),
            mailfrom,
            rcpttos,
            data.decode("utf-8"),
            mail_options,
            rcpt_options,
            "PENDING",
            None,
            None,
        )

        self.database.insert_message(message)



# MIT License
#
# Copyright (c) 2023 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import requests

from pinkman.module.database import Database


class Backend:
    """Backend Class"""

    def __init__(self):
        pass

    def log(self, message):
        print(
            "New message received with id {} and from {}".format(
                message.id, message.mailfrom
            )
        )

        return True

    def http(self, remote, message):
        data = {
            "id": message.id,
            "mailfrom": message.mailfrom,
            "rcpttos": message.rcpttos,
            "data": message.data,
            "mailOptions": message.mail_options,
            "rcptOptions": message.rcpt_options,
            "createdAt": message.created_at,
            "updatedAt": message.updated_at,
        }

        headers = {
            "Content-Type": "application/json",
            "X-API-Key": remote.apikey,
        }

        if remote.method.upper() == "PUT":
            try:
                response = requests.put(remote.url, headers=headers, json=data)
            except Exception as e:
                raise Exception("Remote backend responds with error: {}".format(str(e)))

            if response.status_code // 100 != 2:
                raise Exception(
                    "Remote backend responds with invalid status code {}".format(
                        response.status_code
                    )
                )

        else:
            try:
                response = requests.post(remote.url, headers=headers, json=data)
            except Exception as e:
                raise Exception("Remote backend responds with error: {}".format(str(e)))

            if response.status_code // 100 != 2:
                raise Exception(
                    "Remote backend responds with invalid status code {}".format(
                        response.status_code
                    )
                )

        return True



# MIT License
#
# Copyright (c) 2023 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click

from pinkman import __version__
from pinkman.command.worker import Worker
from pinkman.command.server import Server


@click.group(help="🐺 An SMTP Server to forward Messages into A Backend System")
@click.version_option(version=__version__, help="Show the current version")
def main():
    pass


# Server command
@click.group(help="Server commands")
def server():
    pass


# Run server sub command
@server.command(help="Run the server")
@click.option(
    "-c",
    "--config",
    "config",
    type=click.STRING,
    default="",
    help="Server config file",
)
def run(config):
    return Server().run(config)


# Worker command
@click.group(help="Worker commands")
def worker():
    pass


# Run worker sub command
@worker.command(help="Run the worker")
@click.option(
    "-c",
    "--config",
    "config",
    type=click.STRING,
    default="",
    help="Server config file",
)
def run(config):
    return Worker().run(config)


# Register Commands
main.add_command(server)
main.add_command(worker)


if __name__ == "__main__":
    main()



# MIT License
#
# Copyright (c) 2023 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Message:
    """Message Model"""

    def __init__(
        self,
        id,
        mailfrom,
        rcpttos,
        data,
        mail_options,
        rcpt_options,
        status,
        created_at,
        updated_at,
    ):
        self._id = id
        self._mailfrom = mailfrom
        self._rcpttos = rcpttos
        self._data = data
        self._mail_options = mail_options
        self._rcpt_options = rcpt_options
        self._status = status
        self._created_at = created_at
        self._updated_at = updated_at

    @property
    def id(self):
        return self._id

    @property
    def status(self):
        return self._status

    @property
    def mailfrom(self):
        return self._mailfrom

    @property
    def rcpttos(self):
        return self._rcpttos

    @property
    def data(self):
        return self._data

    @property
    def mail_options(self):
        return self._mail_options

    @property
    def rcpt_options(self):
        return self._rcpt_options

    @property
    def created_at(self):
        return self._created_at

    @property
    def updated_at(self):
        return self._updated_at

    @status.setter
    def status(self, status):
        self._status = status

    @mailfrom.setter
    def mailfrom(self, mailfrom):
        self._mailfrom = mailfrom

    @rcpttos.setter
    def rcpttos(self, rcpttos):
        self._rcpttos = rcpttos

    @data.setter
    def data(self, data):
        self._data = data

    @mail_options.setter
    def mail_options(self, mail_options):
        self._mail_options = mail_options

    @rcpt_options.setter
    def rcpt_options(self, rcpt_options):
        self._rcpt_options = rcpt_options



# MIT License
#
# Copyright (c) 2023 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click
import time

from pinkman.model.remote import Remote
from pinkman.model.message import Message
from pinkman.module.file import File
from pinkman.module.config import Config
from pinkman.module.database import Database
from pinkman.module.backend import Backend


class Worker:
    """Worker Command Class"""

    def __init__(self):
        self.file = File()
        self.config = Config()
        self.backend = Backend()
        self.database = Database()

    def run(self, config_path):
        self.config.load(config_path)

        configs = self.config.get_configs()

        if not self.file.exists(configs["cache"]["path"]):
            self.database.connect(configs["cache"]["path"])
            self.database.migrate()
        else:
            self.database.connect(configs["cache"]["path"])

        remote = Remote(
            configs["backend"]["method"],
            configs["backend"]["url"],
            configs["backend"]["apikey"],
        )

        click.echo("Worker is up and running!")

        while True:
            messages = self.database.filter_messages("PENDING", 10)

            for message in messages:
                if configs["backend"]["type"] == "http":
                    self.backend.http(remote, message)
                    self.database.update_message_status(message.id, "DELIVERED")
                else:
                    self.backend.log(message)
                    self.database.update_message_status(message.id, "DELIVERED")

            time.sleep(1)



# MIT License
#
# Copyright (c) 2023 Clivern
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import sqlite3

from pinkman.model.message import Message


class Database:
    """Database Class"""

    def connect(self, path):
        self.path = path

        self._connection = sqlite3.connect(self.path)

        return self._connection.total_changes

    def migrate(self):
        cursor = self._connection.cursor()

        cursor.execute(
            "CREATE TABLE IF NOT EXISTS mailbox (id TEXT, message TEXT, status TEXT, createdAt TEXT, updatedAt TEXT)"
        )

        cursor.close()

        self._connection.commit()

    def insert_message(self, message):
        cursor = self._connection.cursor()

        result = cursor.execute(
            "INSERT INTO mailbox VALUES ('{}', '{}', '{}', datetime('now'), datetime('now'))".format(
                message.id,
                json.dumps(
                    {
                        "mailfrom": message.mailfrom,
                        "rcpttos": message.rcpttos,
                        "data": message.data,
                        "mail_options": message.mail_options,
                        "rcpt_options": message.rcpt_options,
                    }
                ),
                message.status,
            )
        )

        cursor.close()

        self._connection.commit()

        return result.rowcount

    def update_message_status(self, id, status):
        cursor = self._connection.cursor()

        result = cursor.execute(
            "UPDATE mailbox SET status = '{}' WHERE id = '{}'".format(status, id)
        )

        cursor.close()

        self._connection.commit()

        return result.rowcount

    def filter_messages(self, status, limit=10):
        result = []

        cursor = self._connection.cursor()

        rows = cursor.execute(
            "SELECT id, message, status, createdAt, updatedAt FROM mailbox WHERE status = '{}' LIMIT {}".format(
                status, limit
            )
        ).fetchall()

        cursor.close()

        for row in rows:
            data = json.loads(row[1])

            message = Message(
                row[0],
                data["mailfrom"],
                data["rcpttos"],
                data["data"],
                data["mail_options"],
                data["rcpt_options"],
                row[2],
                row[3],
                row[4],
            )

            result.append(message)

        return result

    def delete_message(self, id):
        cursor = self._connection.cursor()

        cursor.execute("DELETE FROM mailbox WHERE id = ?", (id,))

        cursor.close()

        self._connection.commit()
